# -*- coding: utf-8 -*-
import bpy

import Spr

from mathutils import Vector,Quaternion


import spr_op.creature as creature

import spr_op.visualizer as visualizer

class PhysicsController():
    def __init__(self):
        
        #Springhead2
        self.phSdk = None
        self.crSdk = None
        self.phScene = None
        
        #controllers
        self.solid_controllers = {}
        self.creature_controllers = {}
        self.creatures_num = 0
        
        #temp lists
        self.compound_objects = []
        self.ikeff_objects = []
        self.ika_objects = []
        
        self.prev_time = 0
        
        #info
        self.info_fps_ave = 0
        self.info_fps_list = []
        self.info = {}
        
        self.slow_step_counter = 0
        
        self.record_keyframe_step_counter = 0
        
        #visualizer
        self.visualizer = None
        
        #run scripts
        self.scripts={}
        
        
    def initialize(self):
        
        #initialize Springhead2
        self.phSdk = None
        self.crSdk = None
    
        self.phScene = None
        self.solid_controllers = {}
        self.creature_controllers = {}
        self.creatures_num = 0
        
        self.visualizer_controller = None
        
        #------------------------------
        self.phSdk   = Spr.PHSdk().CreateSdk()
        self.crSdk = Spr.CRSdk().CreateSdk()
        
        
        self.phScene = self.phSdk.CreateScene(Spr.PHSceneDesc())
        g = bpy.context.scene.gravity
        self.phScene.SetGravity(Spr.Vec3f(g.x,g.y,g.z)) #Blenderの座標系Z-upに合わせる
        
        self.visualizer = visualizer.Visualizer(self)
        
        print("Physics Controller: Initialized")
    
    def create_shape(self,obj):
        mesh = None
        
        if(obj.spr_shape_type == "Sphere"):
            sphereDesc = Spr.CDSphereDesc()
            dim = obj.dimensions
            radius = (dim.x+dim.y+dim.z) / 3 / 2
            sphereDesc.radius = float(radius)
            #sphereDesc.radius = float(obj.game.radius)
            mesh = self.phSdk.CreateShape(Spr.CDSphere().GetIfInfoStatic(), sphereDesc)
            
        elif(obj.spr_shape_type == "Box"):
            boxDesc = Spr.CDBoxDesc()
            dim = obj.dimensions
            boxDesc.boxsize = Spr.Vec3d(dim.x,dim.y,dim.z)
            
            mesh = self.phSdk.CreateShape(Spr.CDBox().GetIfInfoStatic(), boxDesc)
        
        elif(obj.spr_shape_type == "Convex"):
            convexDesc = Spr.CDConvexMeshDesc()
            
            v_vec3d = []
            for v in obj.data.vertices:
                scale = obj.scale
                v_vec3d.append(Spr.Vec3d(v.co.x * scale.x, v.co.y * scale.y, v.co.z * scale.z))
            
            convexDesc.vertices = v_vec3d
            
            mesh = self.phSdk.CreateShape(Spr.CDConvexMesh().GetIfInfoStatic(), convexDesc)
        
        
        
        return mesh
        
    
    
    def create_solid(self,obj):
        name = obj.name
        
        #登録済みなら古いsolidをfreeze、辞書から削除
        if(name in self.solid_controllers):
            so_con = self.solid_controllers[name]
            freeze_solid(so_con, self.phScene)
            del self.solid_controllers[name]
        
        if(obj.spr_shape_type!="None"):    #Physics Modelを作る条件
            
            #compound child
            if(obj.spr_compound_child):
                self.compound_objects.append(obj)
                return()
            
            so_con = SolidController()
            so_con.ph_con = self
            so_con.object = obj
            
            #----Create Solid----
            soDesc = Spr.PHSolidDesc()
            so_con.phSolid = self.phScene.CreateSolid(soDesc)
            so_con.phSolid.SetName("so" + name)
            
            #初期姿勢
            loc = obj.location
            rot = obj.rotation_quaternion #(w,x,y,z)
            so_con.phSolid.SetFramePosition(Spr.Vec3d(loc.x,loc.y,loc.z))
            so_con.phSolid.SetOrientation(Spr.Quaterniond(rot.w,rot.x,rot.y,rot.z))
            #initial
            so_con.initial_pos = Vector((loc.x,loc.y,loc.z))
            so_con.initial_rot = Quaternion((rot.w,rot.x,rot.y,rot.z))
            #prev
            so_con.prev_pos = Vector((loc.x,loc.y,loc.z))
            so_con.prev_rot = Quaternion((rot.w,rot.x,rot.y,rot.z))
            
            
            #Create Shape
            mesh = self.create_shape(obj)
            
            #solidにmeshを追加
            if(not mesh is None):
                so_con.phSolid.AddShape(mesh)
            
            #パラメータ適用
            so_con.set_solid_parameters()
            so_con.set_shape_parameters()
            
            #辞書に登録
            self.solid_controllers[name] = so_con
            
            
            #IK PHIKEndEffector----------------------
            if(obj.spr_ik_enabled == 1 and obj.spr_ik_end_effector == 1):
                ikeffDesc = Spr.PHIKEndEffectorDesc()
                ikeffDesc.bEnabled = True
                
                ikeff = self.phScene.CreateIKEndEffector(ikeffDesc)
                #ikeff = self.phScene.CreateIKEndEffector(Spr.PHIKEndEffector().GetIfInfoStatic(),ikeffDesc)
                
                ikeff.SetName("ikeff"+name)
                ikeff.AddChildObject(so_con.phSolid)
                
                so_con.ikeff = ikeff
                
                #パラメータ適用
                so_con.set_ikeff_parameters()
                
                #IKActuatorに後で登録（一時的なリスト）
                self.ikeff_objects.append(obj)
            
            
            #creature-------------------
            if obj.spr_creature_body == 1:
                if obj.spr_creature_id + 1 > self.creatures_num:
                    self.creatures_num = obj.spr_creature_id + 1     #creatureの数を更新
            
    
    def set_compound_objects(self):
        
        parent_so_con_list = [] #パラメータ再適用のため(inertia)
        
        for obj in self.compound_objects:
            parent_obj = obj.parent
            if(not parent_obj is None):
                mesh = self.create_shape(obj)
                
                if(not mesh is None):
                    parent_so_con = self.solid_controllers[parent_obj.name]
                    parent_so_con.phSolid.AddShape(mesh)
                    
                    if not parent_so_con in parent_so_con_list:
                        parent_so_con_list.append(parent_so_con)
        
        for parent_so_con in parent_so_con_list:
            parent_so_con.set_solid_parameters()
        
        del self.compound_objects[:]
        
    
    def create_joint(self,so_con):
        
        jo_objs =[]
        jo_objs.append(so_con.object)
        
        for child_obj in so_con.object.children:
            if child_obj.spr_additional_joint==1:
                jo_objs.append(child_obj)
        
        for jo_obj in jo_objs:
            
            if(jo_obj.spr_joint_type == "None"):
                continue
            
            joDesc = None
            #bjDesc = Spr.PHBallJointDesc()
            #hjDesc = Spr.PHHingeJointDesc()
            
            #対象Solid取得
            #子
            so_child = so_con.phSolid
            #親
            if(jo_obj.spr_joint_target in self.solid_controllers):
                so_parent = self.solid_controllers[jo_obj.spr_joint_target].phSolid
                if(so_child == so_parent):
                    print("Error: Joint Target of " + jo_obj.name + " is Invalid (same object)")
                    return()
            else:
                print("Error: Joint Target of " + jo_obj.name + " is Invalid")
                return()
            
            #衝突判定の設定
            if(jo_obj.spr_joint_collision==0):
                self.phScene.SetContactMode(so_child, so_parent, 0) #ContactMode 0:MODE_NONE 1:MODE_PENALTY 2:MODE_LCP
            else:
                self.phScene.SetContactMode(so_child, so_parent, 2)
            
                
            #Joint作成-----------------------------
            #Joint Desc作成
            if(jo_obj.spr_joint_type=="Ball"):
                joDesc = Spr.PHBallJointDesc()
                #if(obj.spr_limit_swing_enabled==1):
                #     joDesc.limitSwing = Spr.Vec2d(obj.spr_limit_swing_min, obj.spr_limit_swing_max) #min,Max
                #if(obj.spr_limit_twist_enabled==1):
                #     joDesc.limitTwist = Spr.Vec2d(obj.spr_limit_twist_min, obj.spr_limit_twist_max) #min,Max
            if(jo_obj.spr_joint_type=="Hinge"):
                joDesc = Spr.PHHingeJointDesc()
                #if(obj.spr_limit_range_enabled==1):
                #     joDesc.lower = obj.spr_limit_range_min
                #     joDesc.upper = obj.spr_limit_range_max
            if(jo_obj.spr_joint_type=="Slider"):
                joDesc = Spr.PHSliderJointDesc()
                #if(obj.spr_limit_range_enabled==1):
                #     joDesc.lower = obj.spr_limit_range_min
                #     joDesc.upper = obj.spr_limit_range_max
            if(jo_obj.spr_joint_type=="Spring"):
                joDesc = Spr.PHSpringDesc()
                
            
            
            #joint 作成処理
            if(jo_obj.spr_joint_type=="Ball"):
                phJoint = self.phScene.CreateJoint(so_parent, so_child,Spr.PHBallJoint.GetIfInfoStatic() , joDesc)
            if(jo_obj.spr_joint_type=="Hinge"):
                phJoint = self.phScene.CreateJoint(so_parent, so_child,Spr.PHHingeJoint.GetIfInfoStatic() , joDesc)
            if(jo_obj.spr_joint_type=="Slider"):
                phJoint = self.phScene.CreateJoint(so_parent, so_child,Spr.PHSliderJoint.GetIfInfoStatic() , joDesc)
            if(jo_obj.spr_joint_type=="Spring"):
                phJoint = self.phScene.CreateJoint(so_parent, so_child,Spr.PHSpring.GetIfInfoStatic() , joDesc)
            
            ##名前
            if(jo_obj.spr_additional_joint==0):
                phJoint.SetName("jo"+jo_obj.name)
            else:
                phJoint.SetName(jo_obj.name)
            
            ##JointController
            jo_con = JointController()
            jo_con.so_con = so_con
            jo_con.jo_object = jo_obj
            jo_con.phJoint = phJoint
            
            so_con.joint_controllers[jo_obj.name] = jo_con
            
            #Limit
            if isinstance(phJoint, Spr.PHBallJoint):
                limitDesc = Spr.PHBallJointConeLimitDesc()
                phJoint.CreateLimit(Spr.PHBallJointConeLimit.GetIfInfoStatic(), limitDesc)
            if isinstance(phJoint, Spr.PHHingeJoint) or isinstance(phJoint, Spr.PHSliderJoint):
                limitDesc = Spr.PH1DJointLimitDesc()
                phJoint.CreateLimit(limitDesc)
            
            #パラメータ適用
            ##Plug Soclet
            jo_con.set_joint_plug_socket()
            
            ##その他
            jo_con.set_joint_parameters()
            
            
            if(jo_obj == so_con.object):#第1Jointの場合
                #IK PHIKActuator=========================================
                if(jo_obj.spr_ik_enabled==0):
                    continue
                
                ikaDesc = None
                
                if(jo_obj.spr_joint_type=="Ball"):
                    ikaDesc = Spr.PHIKBallActuatorDesc()
                elif(jo_obj.spr_joint_type=="Hinge"):
                    ikaDesc = Spr.PHIKHingeActuatorDesc()
                else:
                    continue
                
                ikaDesc.bEnabled = True
                
                
                ika = None
                if(jo_obj.spr_joint_type=="Ball"):
                    ika = self.phScene.CreateIKActuator(Spr.PHIKBallActuator().GetIfInfoStatic(),ikaDesc)
                elif(jo_obj.spr_joint_type=="Hinge"):
                    ika = self.phScene.CreateIKActuator(Spr.PHIKHingeActuator().GetIfInfoStatic(),ikaDesc)
                else:
                    continue
                
                ika.SetName("ika"+jo_obj.name)
                ika.AddChildObject(phJoint)
                
                #debug
                #print(ika.GetName())
                
                so_con.ika = ika
                
                #パラメータ適用
                so_con.set_ika_parameters()
                
                #後で木構造作成、EndEffector登録（一時的なリスト）
                self.ika_objects.append(jo_obj)
        
        
        
        
    def set_IK(self):#IKのAPI変更後はここでActuatorの構造構築とEndEffectorの登録を行う。
        for obj in self.ika_objects:#IKActuator
            so_con_child = self.solid_controllers[obj.name]
            if obj.spr_joint_target in self.solid_controllers:
                so_con_parent = self.solid_controllers[obj.spr_joint_target]
                
                if so_con_parent.ika is not None:
                    so_con_parent.ika.AddChildObject(so_con_child.ika)
                
            
        for obj in self.ikeff_objects:#IK EndEffector
            so_con = self.solid_controllers[obj.name]
            if so_con.ika is not None:
                so_con.ika.AddChildObject(so_con.ikeff)
        
        del self.ika_objects[:]
        del self.ikeff_objects[:]
    
    
    
    def construct_scene(self):#全てのオブジェクトを追加、joint作成、creature構築
        
        for obj in bpy.data.objects:
            self.create_solid(obj)
        
        self.set_compound_objects()
        
        for name,so_con in self.solid_controllers.items():
            self.create_joint(so_con)
        
        #IK
        self.set_IK()
        
        #creature
        if(bpy.context.scene.spr_creature_enabled == 1):
            i = 0
            while i < self.creatures_num :
                cr_con = creature.CreatureController(self,i)
                self.creature_controllers[i] = cr_con
                cr_con.create_creature()
                
                i += 1
        
        #run script
        self.update_scripts()
        
        
    def add_object(self):#選択オブジェクトを追加、joint作成
        
        for obj in bpy.context.selected_objects:
            self.create_solid(obj)
        
        self.set_compound_objects()
        
        for obj in bpy.context.selected_objects:
            if obj.name in self.solid_controllers:
                so_con = self.solid_controllers[obj.name]
                self.create_joint(so_con)
        
        #IK
        self.set_IK()
        
    def step(self):#Springhead2のステップ
        #g = bpy.context.scene.gravity
        #self.phScene.SetGravity(Spr.Vec3f(g.x,g.y,g.z))#重力適用
        
        if(not bpy.context.scene.spr_creature_enabled):
            self.phScene.Step()
        
        else:#creature
            
            self.phScene.ClearForce()
            self.phScene.GenerateForce()
            
            self.crSdk.Step()
            
            
            """comment
            #koguma
            for id,cr_con in self.creature_controllers.items():
                if(cr_con.ready == True):
                    cr_con.koguma.step()
            comment"""
            
            
            
            self.phScene.Integrate()
            
            #メインシーン以外のステップ
            i = 1
            while i < self.phSdk.NScene():
                #self.phSdk.GetScene(i).SetContactMode(0)#ContactMode 0:MODE_NONE 1:MODE_PENALTY 2:MODE_LCP
                self.phSdk.GetScene(i).Step()
                i += 1
                



    
    def sync(self):#Blender-Springhead2 相互同期処理
        if(self.solid_controllers!={}):
            for name,so_con in self.solid_controllers.items():
                so_con.sync()
        
        #Creature パラメータリアルタイム編集モード
        if(bpy.context.scene.spr_creature_enabled == 1):
            for id,cr_con in self.creature_controllers.items():
                if cr_con.ready == True:
                    if cr_con.core_object.spr_creature_realtime_editing == 1:
                        cr_con.update_parameters()


    
    
#    def set_time_step(self,time_step):
#        self.phScene.SetTimeStep(time_step)
        
    def set_scene_props(self,time_step):
        i = 0
        while i < self.phSdk.NScene():
            scene = self.phSdk.GetScene(i)
            g = bpy.context.scene.gravity
            self.phScene.SetNumIteration(bpy.context.scene.spr_num_iteration)
            self.phScene.SetGravity(Spr.Vec3f(g.x,g.y,g.z))#重力適用
            self.phScene.SetTimeStep(time_step)
            
            #IK
            ikEngine = self.phScene.GetIKEngine()
            ikEngine.Enable(bpy.context.scene.spr_ik_engine_enabled)
            ikEngine.SetNumIter(bpy.context.scene.spr_ik_iteration_num)
            i += 1
        
            
    def restore(self):
        for name,so_con in self.solid_controllers.items():
            so_con.restore()
    
    def save_spr(self):
        fiSdk = Spr.FISdk().CreateSdk()
        fileSpr = fiSdk.CreateFileSpr()
        objs = Spr.ObjectIfs()
        objs.Push(self.phSdk)
        fileSpr.Save(objs,"PHScene.spr")
        
        print("spr file saved as 'PHScene.spr'")
    
    def run_scripts(self, i_substep, substep):#スクリプト実行機能
        for key,script in sorted(self.scripts.items()):
            try:
                exec(script)
            except Exception as exception:#例外情報の記録、スクリプト実行の停止
                self.info["error"]="" #空にする
                self.info["error"] +="Error in running user scripts.\n\n"
                self.info["error"] += "Script Name: " + key +"\n"
                self.info["error"] += "Type: " + str(type(exception)) +"\n"
                self.info["error"] += "Args: " + str(exception.args) +"\n"
                #self.info["error"] += "Message: " + exception.message
                self.info["error"] += "Exception: " + str(exception) +"\n"
                
                print("Error in running user scripts")
                print("Error message is in Text \"spr_error\"")
                
                bpy.context.scene.spr_run_scripts_enabled = 0
                print("Stopped Running scripts")
                
                return()
    
    def update_scripts(self):#スクリプト更新
        #初期化
        self.scripts.clear()
        
        try:#スクリプトが既存ならそちらを使う
            script_text = bpy.data.texts["spr_script"]
        except:#無ければデフォルトの物を作る
            script_text = bpy.data.texts.new("spr_script")
            #以下デフォルトのスクリプト（サンプル）
            script = "#" + "spr_script" + "\n"
            script += "#" + "This is script for SprBlender" + "\n"
            script += "#" + "To Use Multi Scrips, Create Script-Name-List as \"spr_scripts_list\"" + "\n"
            script += "#" + "------------------------------" + "\n"
            script += "#" + "self: PhysicsController" + "\n"
            script += "#" + "self.phSdk: Springhead2 PHSdk" + "\n"
            script += "#" + "self.phScene: Springhead2 PHScene" + "\n"
            script += "#" + "self.solid_controllers[\"name\"]: SolidController" + "\n"
            script += "#" + "substep: substep" + "\n"
            script += "#" + "i_substep: iteration of substep" + "\n"
            script += "#" + "------------------------------" + "\n"
            
            
            script_text.from_string(script)
            
        self.scripts["spr_script"] = script_text.as_string()
        
        #複数Textに対応
        if("spr_scripts_list" in bpy.data.texts):
            for line in bpy.data.texts["spr_scripts_list"].lines:
                script_name = line.body
                if(script_name in bpy.data.texts):
                    self.scripts[script_name] = bpy.data.texts[script_name].as_string()



class SolidController():
    ph_con = None
    object = None
    phSolid = None
    joint_controllers = {}
    ikeff = None
    ika = None

    prev_pos = Vector((0.0,0.0,0.0)) #Blender Vector
    prev_rot = Quaternion((1.0, 0.0, 0.0, 0.0)) #Blender Quaternion w,x,y,z
    initial_pos = Vector((0.0,0.0,0.0)) #Blender Vector
    initial_rot = Quaternion((1.0, 0.0, 0.0, 0.0)) #Blender Quaternion w,x,y,z
    
    #control--------------------------------------------------------
    def solid_control_apply(self,attr):#int attr 0:velocity 1:ang v 2:force 3:torque
        solid = self.phSolid
        obj = self.object
        if (not obj is None) and (not solid is None):
            if(attr==0):#Velocity
                velocity = Spr.Vec3d(obj.spr_solid_velocity_x, obj.spr_solid_velocity_y, obj.spr_solid_velocity_z)
                if(obj.spr_solid_velocity_local==1):#local
                    velocity = solid.GetPose().getOri() * velocity
                solid.SetVelocity(velocity)
                
            elif(attr==1):#Angular Velocity
                ang_velocity = Spr.Vec3d(obj.spr_solid_angular_velocity_x, obj.spr_solid_angular_velocity_y, obj.spr_solid_angular_velocity_z)
                if(obj.spr_solid_ang_velocity_local==1):#local
                    ang_velocity = solid.GetPose().getOri() * ang_velocity
                solid.SetAngularVelocity(ang_velocity)
                
            elif(attr==2):#Add Force
                force = Spr.Vec3d(obj.spr_solid_force_x, obj.spr_solid_force_y, obj.spr_solid_force_z)
                if(obj.spr_solid_set_force_point==1):#作用点を指定
                    solid.AddForce(force, Spr.Vec3d(obj.spr_solid_force_point_x, obj.spr_solid_force_point_y, obj.spr_solid_force_point_z))
                else:
                    if(obj.spr_solid_force_local==1):#local
                        force = solid.GetPose().getOri() * force
                    solid.AddForce(force)
                    
            elif(attr==3):#Add Torque
                torque = Spr.Vec3d(obj.spr_solid_torque_x, obj.spr_solid_torque_y, obj.spr_solid_torque_z)
                if(obj.spr_solid_torque_local==1):#local
                    torque = solid.GetPose().getOri() * torque
                solid.AddTorque(torque)
                

    
    #set parameters--------------------------------------------------------
    def set_joints_parameters(self):
        if(self.joint_controllers!={}):
            for name, jo_con in self.joint_controllers.items():
                jo_con.set_joint_parameters()
        
        
    def set_ika_parameters(self):
        ika = self.ika
        obj = self.object
        
        if (not obj is None) and (not ika is None):
            ika.SetBias(obj.spr_ik_bias)
            #ika.SetSpring(obj.spr_ik_spring)#IK Spring,Damper廃止
            #ika.SetDamper(obj.spr_ik_damper)
            
        

    def set_ikeff_parameters(self):
        ikeff = self.ikeff
        obj = self.object
        
        if (not obj is None) and (not ikeff is None):
            ikeff.SetTargetPosition(Spr.Vec3d(obj.spr_ik_target_local_pos_x, obj.spr_ik_target_local_pos_y, obj.spr_ik_target_local_pos_z))
            
            #IK Control
            ikeff.EnablePositionControl(obj.spr_ik_pos_control_enabled)
            ikeff.EnableOrientationControl(obj.spr_ik_ori_control_enabled)
            
            if obj.spr_ik_target_object_enabled==1 and (obj.spr_ik_target_object_name in bpy.data.objects):#target objectを目標にする場合
                obj_target = bpy.data.objects[obj.spr_ik_target_object_name]
                #obj_targetの位置・姿勢を取得
                if(obj_target.parent is None):#childでない場合
                    loc = obj_target.location
                    rot = obj_target.rotation_quaternion #(w,x,y,z)
                else:#childの場合(static child)
                    #world座標を取得する
                    mat = obj_target.matrix_world
                    loc = mat.to_translation()
                    rot = mat.to_quaternion()
                #値を設定
                ikeff.SetTargetPosition(Spr.Vec3d(loc.x, loc.y, loc.z))
                ikeff.SetTargetOrientation(Spr.Quaterniond(rot.w,rot.x,rot.y,rot.z))
            
            else:#目標を値で設定する場合
                ikeff.SetTargetPosition(Spr.Vec3d(obj.spr_ik_target_pos_x, obj.spr_ik_target_pos_y, obj.spr_ik_target_pos_z))
                ori = Spr.Quaterniond()
                ####FromEulerを使うときは座標系の違いに注意 Blender x,y,z →Springhead z,x,y
                pi =  3.1415926535
                ori.FromEuler(Spr.Vec3d(obj.spr_ik_target_ori_y*pi/180, obj.spr_ik_target_ori_z*pi/180, obj.spr_ik_target_ori_x*pi/180))
                ikeff.SetTargetOrientation(ori)
                

            
    def set_shape_parameters(self):
        solid = self.phSolid
        obj = self.object
        
        if (not obj is None) and (not solid is None):
            #material
            material = obj.material_slots.id_data.active_material
            if(not material is None):
                mesh = solid.GetShape(0)
                mesh.SetElasticity(material.physics.elasticity)
                mesh.SetDynamicFriction(material.physics.friction)#動摩擦力
                mesh.SetStaticFriction(material.spr_friction_static)#静止摩擦力
    
    def set_solid_parameters(self):
        solid = self.phSolid
        obj = self.object
        
        if (not obj is None) and (not solid is None):
            solid.SetMass(obj.game.mass)
            solid.SetCenterOfMass(Spr.Vec3d(obj.spr_center_x, obj.spr_center_y, obj.spr_center_z))
            
            #Static -> Dynamical False
            if(obj.spr_physics_type=="Static"):
                solid.SetDynamical(False)
            elif(not obj.parent is None):#static child
                solid.SetDynamical(False)
            else:
                solid.SetDynamical(True)
            
            
            #inertia
            if(obj.spr_inertia_manual_setting_enabled == 1):
                #手動入力
                inertia = Spr.Matrix3d()
                inertia.xx = obj.spr_inertia_x
                inertia.yy = obj.spr_inertia_y
                inertia.zz = obj.spr_inertia_z
                solid.SetInertia(inertia)
            elif(obj.spr_shape_type == "Sphere"):
                #自動計算
                dim = obj.dimensions
                radius = (dim.x+dim.y+dim.z) / 3 / 2
                i = obj.game.mass * radius * radius * 2 / 5
                #i = obj.game.mass * obj.game.radius * obj.game.radius * 2 / 5
                inertia = Spr.Matrix3d()
                inertia.xx = i
                inertia.yy = i
                inertia.zz = i
                solid.SetInertia(inertia)
                
            elif(obj.spr_shape_type == "Box"):
                #自動計算
                dim = obj.dimensions
                x = dim.x
                y = dim.y
                z = dim.z
                inertia = Spr.Matrix3d()
                #対角行列をつくる
                inertia.xx = obj.game.mass*(y * y + z * z)/12
                inertia.yy = obj.game.mass*(z * z + x * x)/12
                inertia.zz = obj.game.mass*(x * x + y * y)/12
                solid.SetInertia(inertia)
                
            elif(obj.spr_shape_type == "Convex"):
                #自動計算 全てのshapeを考慮
                volume_list = []
                center_list = []
                inertia_list = []
                
                #shapeから体積、質量中心、質量中心周りの慣性行列(単位密度)を取得
                shapes = solid.GetShapes()
                for shape in shapes:
                    volume_list.append(shape.CalcVolume())
                    center_list.append(shape.CalcCenterOfMass())
                    inertia_list.append(shape.CalcMomentOfInertia())
                    
                #体積の合計
                volume_sum = 0
                for volume in volume_list:
                    volume_sum += volume
                
                #密度
                density = 0
                if(volume_sum!=0):
                    density = obj.game.mass / volume_sum
                    
                    #物体全体の慣性行列を求める
                    inertia = Spr.Matrix3d()
                    inertia *= 0 #単位行列になってるので一旦ゼロにする
                    num_shapes = len(volume_list)
                    
                    i = 0
                    while i < num_shapes:
                        inertia_i = Spr.Matrix3d()
                        inertia_c = inertia_list[i]
                        center = center_list[i]
                        volume = volume_list[i]
                        
                        inertia_i.xx = inertia_c.xx + (center.y * center.y + center.z * center.z) * volume
                        inertia_i.xy = inertia_c.xy - center.x * center.y * volume
                        inertia_i.xz = inertia_c.xz - center.x * center.z * volume
                        inertia_i.yx = inertia_c.yx - center.y * center.x * volume
                        inertia_i.yy = inertia_c.yy + (center.z * center.z + center.x * center.x) * volume
                        inertia_i.yz = inertia_c.yz - center.y * center.z * volume
                        inertia_i.zx = inertia_c.zx - center.z * center.x * volume
                        inertia_i.zy = inertia_c.zy - center.z * center.y * volume
                        inertia_i.zz = inertia_c.zz + (center.x * center.x + center.y * center.y) * volume
                        
                        inertia += inertia_i * density
                        
                        i += 1
                        
                    solid.SetInertia(inertia)
                    
                    #debug out
                    debug = "Inertia of " + obj.name + "\n"
                    debug += "num of shapes: " + str(num_shapes) + "\n"
                    debug += "mass: " + str(obj.game.mass) + "\n"
                    debug += "volume: " + str(volume_sum) + "\n"
                    debug += "density: " + str(density) + "\n"
                    debug += "inertia: " + "\n"
                    debug += str(inertia)
                    
                    self.ph_con.info["inertia_calc_log_"+obj.name] = debug
                    


    #sync-----------------------------------------------------------------------
    def sync(self):
        #obj = bpy.data.objects[self.name]
        obj = self.object
        loc = None
        rot = None

        if(obj.parent is None):#childでない場合
            loc = obj.location
            rot = obj.rotation_quaternion #(w,x,y,z)
        else:#childの場合(static child)
            #world座標を取得する
            mat = obj.matrix_world
            loc = mat.to_translation()
            rot = mat.to_quaternion()
            #scale = mat.to_scale()
        
        if(self.prev_pos != loc or self.prev_rot != rot):#Blender側(GUI)で変位、回転の操作があった場合 Blender -> Springhead2
            #debug
            #print(self.prev_pos," ",loc)
            #print(self.prev_rot," ",rot)
            
            #Springhead2側を同期
            self.phSolid.SetFramePosition(Spr.Vec3d(loc.x,loc.y,loc.z))
            self.phSolid.SetOrientation(Spr.Quaterniond(rot.w,rot.x,rot.y,rot.z))
            
            #回転速度 → ゼロにする
            self.phSolid.SetAngularVelocity(Spr.Vec3d(0.0, 0.0, 0.0))
            
            #速度
            delta = (0,0,0)
            if (bpy.context.scene.spr_throw_enabled == 1 and obj.spr_physics_type != "Static"):#投げモード
                delta = (loc.x - self.prev_pos.x, loc.y - self.prev_pos.y, loc.z - self.prev_pos.z)
                if(delta[0]>1 or delta[1]>1 or delta[2]>1):
                    delta = (0,0,0)
            factor = bpy.context.scene.game_settings.frequency
            v = Spr.Vec3d(delta[0]*factor, delta[1]*factor, delta[2]*factor)
            self.phSolid.SetVelocity(v)
            
            #prev pos,rot更新
            self.prev_pos = Vector((loc.x,loc.y,loc.z))#Vectorをそのまま渡すとprev_posがactiveObj.locationの参照になってしまう
            self.prev_rot = Quaternion((rot.w,rot.x,rot.y,rot.z))
            
            
        else: #Springhead2 -> Blender
            pos = self.phSolid.GetPose().getPos()
            ori = self.phSolid.GetPose().getOri()
            
            #※重要※
            #BlenderとSpringhead2との間で座標系の変換は行なっていない
            #Blenderの座標系のままSpringhead2に渡している
            #qt = Quaterniond(0.707106781,-0.707106781,0,0) #1 / (2^0.5) = 0.707106781 座標系の変換を試みたがうまくいかず
            #ori = ori * qt
            
            if(obj.parent is None):#childでない場合
                obj.location = (pos.x,pos.y,pos.z)
                
                #姿勢適用・・・rotation_modeを毎回変更してるとすごく重くなってしまう→現状QUATERNIONのまま渡すしか無い
                ##QUATERNION
                if(obj.rotation_mode != "QUATERNION"):
                    obj.rotation_mode = "QUATERNION"
                obj.rotation_quaternion = Quaternion((ori.w,ori.x,ori.y,ori.z))
                
                ##Euler・・・Springhead2側でToEulerするのもやはり重くなってしまう
                #euler = Spr.Vec3d()
                #ori.ToEuler(euler)
                #if(obj.rotation_mode != "XYZ"):
                #    obj.rotation_mode = "XYZ"
                #obj.rotation_euler = Vector((euler.x, euler.y, euler.z))
                
                
                #Record KeyFrame
                if(bpy.context.scene.spr_record_keyframe==1 and bpy.context.scene.spr_step_enabled==1 and self.ph_con.record_keyframe_step_counter == 0 and obj.spr_physics_type != "Static"):
                    if(bpy.context.scene.spr_use_unique_keyframe == 1):#独自のキーフレーム番号で記録する場合
                        obj.keyframe_insert(data_path="location", frame = bpy.context.scene.spr_unique_keyframe)
                        if(obj.rotation_mode != "QUATERNION"):
                            obj.rotation_mode = "QUATERNION"
                        obj.keyframe_insert(data_path="rotation_quaternion", frame = bpy.context.scene.spr_unique_keyframe)
                    else:#シーンのカレントキーフレームに記録する場合
                        obj.keyframe_insert(data_path="location")
                        if(obj.rotation_mode != "QUATERNION"):
                            obj.rotation_mode = "QUATERNION"
                        obj.keyframe_insert(data_path="rotation_quaternion")
                    
                
            #prev pos,rot更新
            self.prev_pos = Vector((pos.x,pos.y,pos.z))
            self.prev_rot = Quaternion((ori.w,ori.x,ori.y,ori.z))
        
        #パラメータリアルタイム編集モード
        if(obj.spr_solid_realtime_editing==1):
            self.set_solid_parameters()
            self.set_shape_parameters()
        
        if(self.joint_controllers!={}):
            for name, jo_con in self.joint_controllers.items():
                if jo_con.jo_object.spr_joint_realtime_editing==1:
                    jo_con.set_joint_parameters()
        
        if(obj.spr_ik_realtime_editing==1):
            self.set_ika_parameters()
            self.set_ikeff_parameters()
        
        #Solid realtime control
        if(obj.spr_solid_velocity_realtime_control_enabled==1):
            self.solid_control_apply(0)
        if(obj.spr_solid_ang_velocity_realtime_control_enabled==1):
            self.solid_control_apply(1)
        if(obj.spr_solid_force_realtime_control_enabled==1):
            self.solid_control_apply(2)
        if(obj.spr_solid_torque_realtime_control_enabled==1):
            self.solid_control_apply(3)
        
        
    
    def restore(self):
        loc = self.initial_pos
        rot = self.initial_rot
        
        self.phSolid.SetFramePosition(Spr.Vec3d(loc.x,loc.y,loc.z))
        self.phSolid.SetOrientation(Spr.Quaterniond(rot.w,rot.x,rot.y,rot.z))
        self.phSolid.SetAngularVelocity(Spr.Vec3d(0.0, 0.0, 0.0))
        self.phSolid.SetVelocity(Spr.Vec3d(0.0,0.0,0.0))
        

class JointController():
    so_con = None
    jo_object = None
    phJoint = None
    
    
    def set_joint_plug_socket(self):
        joint = self.phJoint
        jo_obj = self.jo_object
        
        if (not jo_obj is None) and (not joint is None):
            #代入用変数初期化
            plug = Spr.Vec3d()
            plug_ori = Spr.Quaterniond()
            socket = Spr.Vec3d()
            socket_ori = Spr.Quaterniond()
            
            ##plug:子オブジェクトから見た位置・姿勢
            if(jo_obj.spr_additional_joint==0):#追加Jointじゃない場合
                plug = Spr.Vec3d(jo_obj.spr_joint_plug_pos_x, jo_obj.spr_joint_plug_pos_y, jo_obj.spr_joint_plug_pos_z)
                plug_ori = Spr.Quaterniond(jo_obj.spr_joint_plug_ori_w, jo_obj.spr_joint_plug_ori_x, jo_obj.spr_joint_plug_ori_y, jo_obj.spr_joint_plug_ori_z)
            else:#追加Jointの場合
                loc = jo_obj.location
                rot = jo_obj.rotation_quaternion
                plug = Spr.Vec3d(loc.x, loc.y, loc.z)
                plug_ori = Spr.Quaterniond(rot.w, rot.x, rot.y, rot.z)
            
            ###plug offset
            plug_offset = Spr.Quaterniond()
            ####FromEulerを使うときは座標系の違いに注意 Blender x,y,z →Springhead z,x,y
            pi =  3.1415926535
            plug_offset.FromEuler(Spr.Vec3d(jo_obj.spr_joint_plug_offset_y*pi/180, jo_obj.spr_joint_plug_offset_z*pi/180, jo_obj.spr_joint_plug_offset_x*pi/180))
            
            #SetPlugPose
            joint.SetPlugPose(Spr.Posed(plug, plug_ori*plug_offset))
            
            
            ##socket:親から見たsocketの位置・姿勢
            if jo_obj.spr_joint_socket_enabled == 1:
                ###plugとは別にsocketを設定する場合
                socket = Spr.Vec3d(jo_obj.spr_joint_socket_pos_x, jo_obj.spr_joint_socket_pos_y, jo_obj.spr_joint_socket_pos_z)
                socket_ori = Spr.Quaterniond(jo_obj.spr_joint_socket_ori_w, jo_obj.spr_joint_socket_ori_x, jo_obj.spr_joint_socket_ori_y, jo_obj.spr_joint_socket_ori_z)
                
            else:
                ###plugの位置・姿勢を用いる場合
                pose_child = self.so_con.phSolid.GetPose()
                pose_parent = joint.GetSocketSolid().GetPose()
                
                ###pivotのworld座標を計算
                mat_child = Spr.Matrix3d()
                pose_child.getOri().ToMatrix(mat_child)
                plug_w = pose_child.getPos() + mat_child * plug
                
                plug_g2 = plug_w - pose_parent.getPos()
                
                ###socket:親から見たpivotの位置を計算
                mat_parent = Spr.Matrix3d()
                pose_parent.getOri().Inv().ToMatrix(mat_parent)
                socket = mat_parent * plug_g2
                
                ###socket_ori:親から見たpivotの姿勢を計算
                socket_ori =  pose_parent.getOri().Inv() * (pose_child.getOri() * plug_ori)
                
            
            #SetSocketPose
            joint.SetSocketPose(Spr.Posed(socket, socket_ori))
            
    
    def set_joint_parameters(self):
        joint = self.phJoint
        jo_obj = self.jo_object
        
        if (not jo_obj is None) and (not joint is None):
            #有効・無効
            joint.Enable(jo_obj.spr_joint_enabled)
            
            if isinstance(joint, Spr.PHSpring):#PHSpring
                joint.SetSpring(Spr.Vec3d(jo_obj.spr_spring_x, jo_obj.spr_spring_y, jo_obj.spr_spring_z))
                joint.SetDamper(Spr.Vec3d(jo_obj.spr_damper_x, jo_obj.spr_damper_y, jo_obj.spr_damper_z))
                joint.SetSpringOri(jo_obj.spr_spring_ori)
                joint.SetDamperOri(jo_obj.spr_damper_ori)
            
            else:#その他
                joint.SetSpring(jo_obj.spr_spring)
                joint.SetDamper(jo_obj.spr_damper)
                
                #Limit
                limit = joint.GetLimit()
                limit.SetSpring(jo_obj.spr_limit_spring)
                limit.SetDamper(jo_obj.spr_limit_damper)
                if isinstance(joint, Spr.PHBallJoint):
                    limit.SetSwingRange(Spr.Vec2d(jo_obj.spr_limit_swing_min, jo_obj.spr_limit_swing_max))
                    limit.SetTwistRange(Spr.Vec2d(jo_obj.spr_limit_twist_min, jo_obj.spr_limit_twist_max))
                if isinstance(joint, Spr.PHHingeJoint) or isinstance(joint, Spr.PHSliderJoint):
                    limit.SetRange(Spr.Vec2d(jo_obj.spr_limit_range_min, jo_obj.spr_limit_range_max))
            
            #Deform Type->廃止
            #if(obj.spr_joint_deform_type != "Elastic"):
            #    if(obj.spr_joint_deform_type != "Plastic"):
            #        joint.SetDefomationType(1)
            #    elif(obj.spr_joint_deform_type != "Elastic-Plastic"):
            #        joint.SetDefomationType(2)
            #    
            #    joint.SetSecondDamper(obj.spr_second_damper)
            #    joint.SetYieldStress(obj.spr_yield_stress)
            #    joint.SetHardnessRate(obj.spr_hardness_rate)
            #else:
            #    joint.SetDefomationType(0) #Elastic
            
            if isinstance(joint, Spr.PHBallJoint) or isinstance(joint, Spr.PHSpring):
                joint.SetSecondDamper(Spr.Vec3d(jo_obj.spr_second_damper_x, jo_obj.spr_second_damper_y, jo_obj.spr_second_damper_z))
            else:
                joint.SetSecondDamper(jo_obj.spr_second_damper)
            joint.SetYieldStress(jo_obj.spr_yield_stress)
            joint.SetHardnessRate(jo_obj.spr_hardness_rate)
            
            
            #Joint Control
            if isinstance(joint, Spr.PHHingeJoint):#1自由度関節Hinge
                pi =  3.1415926535
                joint.SetTargetPosition(jo_obj.spr_joint_target_position*pi/180)
                joint.SetTargetVelocity(jo_obj.spr_joint_target_velocity)
                joint.SetOffsetForce(jo_obj.spr_joint_offset_force)
                joint.SetMaxForce(jo_obj.spr_joint_max_force)
                
            elif isinstance(joint, Spr.PHSliderJoint):#1自由度関節Slider
                joint.SetTargetPosition(jo_obj.spr_joint_target_position)
                joint.SetTargetVelocity(jo_obj.spr_joint_target_velocity)
                joint.SetOffsetForce(jo_obj.spr_joint_offset_force)
                joint.SetMaxForce(jo_obj.spr_joint_max_force)
                
            elif isinstance(joint, Spr.PHBallJoint):#Ball Joint
                ori = Spr.Quaterniond()
                ####FromEulerを使うときは座標系の違いに注意 Blender x,y,z →Springhead z,x,y
                pi =  3.1415926535
                ori.FromEuler(Spr.Vec3d(jo_obj.spr_joint_target_position_y*pi/180, jo_obj.spr_joint_target_position_z*pi/180, jo_obj.spr_joint_target_position_x*pi/180))
                joint.SetTargetPosition(ori)
                joint.SetTargetVelocity(Spr.Vec3d(jo_obj.spr_joint_target_velocity_x, jo_obj.spr_joint_target_velocity_y, jo_obj.spr_joint_target_velocity_z))
                joint.SetOffsetForce(Spr.Vec3d(jo_obj.spr_joint_offset_force_x, jo_obj.spr_joint_offset_force_y, jo_obj.spr_joint_offset_force_z))
                joint.SetMaxForce(jo_obj.spr_joint_max_force)


def freeze_solid(so_con,phScene):
    
    #if(not so_con.phJoint is None):
    #    #Jointを削除あるいは無効化あるいはchild削除しなければならないが現在いずれもできない？
    #    so_con.phJoint.SetName("none")
    #    so_con.phJoint.Enable(False)
    if(not so_con.ikeff is None):
        so_con.ikeff.SetName("none")
    if(not so_con.ika is None):
        so_con.ika.SetName("none")
    
    so_con.phSolid.SetName("none")
    so_con.phSolid.SetFramePosition(Spr.Vec3d(-10**16,-10**16,-10**16))#-1光年
    so_con.phSolid.SetDynamical(0)
    phScene.SetContactMode(so_con.phSolid,0)
    #phScene.DelChildObject(so_con.phSolid)

